==> quadrant.transit-0.8.12.tgz: package/src/lib/SQL.ts <==
/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only. 
 */

import { Command, EntityNamespacer, EntityNameEscaper } from "./Command";

export function insertSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    let params = cmd.spec.operation.parameters;
    let names : string[] = [];
    let values : any[] = [];
    for (var i = 0; i < params.length; i++) {
        if (i < params.length - 1) {
            if (params[i].required || cmd.args[i] != null) {
                names.push(params[i].spec.name);
                values.push(cmd.args[i]);
            }
        } else {
            for (var n in cmd.args[i]) {
                names.push(n);
                values.push(cmd.args[i][n]);
            }
        }
    }

    return `INSERT INTO ${namespacer(cmd.spec.entity)} (${
        serializeSQLNameList(names, escaper)
    }) VALUES (${
        serializeSQLValueList(values)
    })`;
}

export function updateSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    let nameList : string[] = [];
    let columns : { [key: string] : any } = cmd.args[cmd.args.length-1];
    for (let column in columns) {
        nameList.push(`${escaper(column)} = ${serializeJavaScriptLiteralAsSQL(columns[column])}`);
    }

    return `UPDATE ${namespacer(cmd.spec.entity)} SET ${ 
        nameList.join(", ")
    } WHERE ${
        identityAsSQLFilter(cmd, escaper)
    }`;
}
        
export function deleteSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    return `DELETE FROM ${namespacer(cmd.spec.entity)} WHERE ${identityAsSQLFilter(cmd, escaper)}`;
}

export function truncateSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    return `TRUNCATE TABLE ${namespacer(cmd.spec.entity)}`;
}

export function selectSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    let cols    = cmd.args[0];
    let filters = cmd.args[1];
    let sort    = cmd.args[2];
    let filterSQL = '';
    if (filters === null) {
        // do nothing
    } else if (typeof filters == "object") {
        if (!Array.isArray(filters)) {
            filters = Object.keys(filters).map((col) => [col, Array.isArray(filters[col]) ? "in" : "=", filters[col]]);
        }
        filterSQL = filters.map((filt) => {
            if (!Array.isArray(filt)) {
                throw new Error(`Filter must be of form: ["column", "=", "value"], got: ${filt}`);
            }
            let indexOfOp = 
                ["=", "!=", "<>", "<", "<=", ">", ">=", "!<", "!>", "like", "not like", "is", "in"]
                .findIndex((e) => String(e).toLowerCase() == filt[1]);
            if (indexOfOp == -1)
                throw new Error(`Illegal comparator in filter: ${filt[1]}`);
            if (filt[1] == "in" && Array.isArray(filt[2]) && filt[2].length == 0) {
                return `${escaper(filt[0])} = NULL`; // an empty list will result in parse error - always fail this condition
            } else {
                return `${escaper(filt[0])} ${filt[1]} ${serializeJavaScriptLiteralAsSQL(filt[2])}`;
            }
        }).join(' AND ');
    }
    let orderBySQL = sort.map((col) => col.charAt(0) == '!' ? escaper(col.substr(1)) + " DESC" : escaper(col)).join(", ");
    return `SELECT ${ cols.map(arg => escaper(arg)).join(", ") || "*" }
            FROM ${namespacer(cmd.spec.entity)}
            ${ filterSQL ? "WHERE" : "" } ${filterSQL}
            ${ orderBySQL ? "ORDER BY" : "" } ${orderBySQL}`;
}

export function identityAsSQLFilter( cmd : Command, escaper : EntityNameEscaper ) : string {
    return cmd.spec.entity.identity.map((field, i) => {
        if (cmd.args[i] === undefined) {
            throw new Error(`Parameter ${i} (${field.name}) of ${cmd.spec.entity.name}.${cmd.spec.operation.name} is required.`);
        } else {
            return `${escaper(field.name)} = ${serializeJavaScriptLiteralAsSQL(cmd.args[i])}`;
        }
    }).join(" AND ");
}

export function serializeSQLNameList(names : string[], escaper : EntityNameEscaper) : string {
	return names.map(name => escaper(name)).join(", ");
}

export function serializeSQLValueList(args: any[]) : string {
	return args.map(arg => serializeJavaScriptLiteralAsSQL(arg)).join(", ");
}


export function serializeSQLString(value : string) : string {
	return "'" + value.replace(/'/g, "''") + "'"
}

export function serializeJavaScriptLiteralAsSQL(value: any) : string {
    if (value === undefined) value = null;
    if (Array.isArray(value)) {
        return "(" + value.map(v => serializeJavaScriptLiteralAsSQL(v)).join(",") + ")";
    } else if (typeof value == "string") {
		return serializeSQLString(value);
	} else if (typeof value == "boolean") {
        return value ? '1' : '0';
    } else if (typeof value == "object") {
        if (value === null) {
            return "NULL";
        } else {
            throw new Error("unable to serialize a JavaScript object as SQL");
        }
    } else {
		return String(value);
	}
}

==> quadrant.transit-0.8.17.tgz: package/src/lib/SQL.ts <==
/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only. 
 */

import { Command, DataType, EntityNamespacer, EntityNameEscaper } from "./Command";

export function insertSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    let params = cmd.spec.operation.parameters;
    let names : string[] = [];
    let values : any[] = [];
    for (var i = 0; i < params.length; i++) {
        if (params[i].spec.dataType != DataType.Object) {
            if (params[i].required || cmd.args[i] != null) {
                names.push(params[i].spec.name);
                values.push(cmd.args[i]);
            }
        } else {
            for (var n in cmd.args[i]) {
                names.push(n);
                values.push(cmd.args[i][n]);
            }
        }
    }

    return `INSERT INTO ${namespacer(cmd.spec.entity)} (${
        serializeSQLNameList(names, escaper)
    }) VALUES (${
        serializeSQLValueList(values)
    })`;
}

export function updateSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    let nameList : string[] = [];
    let columns : { [key: string] : any } = cmd.args[cmd.args.length-1];
    for (let column in columns) {
        nameList.push(`${escaper(column)} = ${serializeJavaScriptLiteralAsSQL(columns[column])}`);
    }

    return `UPDATE ${namespacer(cmd.spec.entity)} SET ${ 
        nameList.join(", ")
    } WHERE ${
        identityAsSQLFilter(cmd, escaper)
    }`;
}
        
export function deleteSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    return `DELETE FROM ${namespacer(cmd.spec.entity)} WHERE ${identityAsSQLFilter(cmd, escaper)}`;
}

export function truncateSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    return `TRUNCATE TABLE ${namespacer(cmd.spec.entity)}`;
}

export function selectSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    let cols    = cmd.args[0];
    let filters = cmd.args[1];
    let sort    = cmd.args[2];
    let options = cmd.args[3];
    let filterSQL = serializeFilter(filters, escaper);
    let orderBySQL = sort.map((col) => col.charAt(0) == '!' ? escaper(col.substr(1)) + " DESC" : escaper(col)).join(", ");
    if (options.pageSize) {
        if (options.pageIndex === 0 || options.pageIndex <= 0) {
            throw new Error(`pageIndex must be 1 or greater, received ${options.pageIndex}`)
        }
        orderBySQL += ` OFFSET ${((options.pageIndex||1)-1)*options.pageSize} ROWS FETCH NEXT ${options.pageSize} ROWS ONLY`;
    }
    return `SELECT ${ cols.map(arg => escaper(arg)).join(", ") || "*" }
            FROM ${namespacer(cmd.spec.entity)}
            ${ filterSQL ? "WHERE" : "" } ${filterSQL}
            ${ orderBySQL ? "ORDER BY" : "" } ${orderBySQL}`;
}

export function countSQL( cmd : Command, namespacer : EntityNamespacer, escaper : EntityNameEscaper ) : string {
    let filterSQL = serializeFilter(cmd.args[0], escaper);
    return `SELECT count(*) AS count FROM ${namespacer(cmd.spec.entity)} ${ filterSQL ? "WHERE" : "" } ${filterSQL}`;
}

function serializeFilter(filters, escaper) : string {
    if (filters && typeof filters == "object") {
        if (!Array.isArray(filters)) {
            filters = Object.keys(filters).map((col) => [col, Array.isArray(filters[col]) ? "in" : "=", filters[col]]);
        }
        return filters.map((filt) => {
            if (!Array.isArray(filt)) {
                throw new Error(`Filter must be of form: ["column", "=", "value"], got: ${filt}`);
            }
            let indexOfOp = 
                ["=", "!=", "<>", "<", "<=", ">", ">=", "!<", "!>", "like", "not like", "is", "in"]
                .findIndex((e) => String(e).toLowerCase() == filt[1]);
            if (indexOfOp == -1)
                throw new Error(`Illegal comparator in filter: ${filt[1]}`);
            if (filt[1] == "in" && Array.isArray(filt[2]) && filt[2].length == 0) {
                return `${escaper(filt[0])} = NULL`; // an empty list will result in parse error - always fail this condition
            } else {
                return `${escaper(filt[0])} ${filt[1]} ${serializeJavaScriptLiteralAsSQL(filt[2])}`;
            }
        }).join(' AND ');
    } else {
        return '';
    }
}

export function identityAsSQLFilter( cmd : Command, escaper : EntityNameEscaper ) : string {
    return cmd.spec.entity.identity.map((field, i) => {
        if (cmd.args[i] === undefined) {
            throw new Error(`Parameter ${i} (${field.name}) of ${cmd.spec.entity.name}.${cmd.spec.operation.name} is required.`);
        } else {
            return `${escaper(field.name)} = ${serializeJavaScriptLiteralAsSQL(cmd.args[i])}`;
        }
    }).join(" AND ");
}

export function serializeSQLNameList(names : string[], escaper : EntityNameEscaper) : string {
	return names.map(name => escaper(name)).join(", ");
}

export function serializeSQLValueList(args: any[]) : string {
	return args.map(arg => serializeJavaScriptLiteralAsSQL(arg)).join(", ");
}


export function serializeSQLString(value : string) : string {
	return "'" + value.replace(/'/g, "''") + "'"
}

export function serializeJavaScriptLiteralAsSQL(value: any) : string {
    if (value === undefined) value = null;
    if (Array.isArray(value)) {
        return "(" + value.map(v => serializeJavaScriptLiteralAsSQL(v)).join(",") + ")";
    } else if (typeof value == "string") {
		return serializeSQLString(value);
	} else if (typeof value == "boolean") {
        return value ? '1' : '0';
    } else if (typeof value == "object") {
        if (value === null) {
            return "NULL";
        } else {
            throw new Error("unable to serialize a JavaScript object as SQL");
        }
    } else {
		return String(value);
	}
}
